/* Copyright 2007
   University of Utah  
   Carlos Scheidegger - cscheid@sci.utah.edu

   This code is GPL v2 or higher, your choice.
*/ 

#ifndef TRIMESH_H_INCLUDED
#include <TriMesh.h>
#define TRIMESH_H_INCLUDED
#endif

#ifndef IOSTREAM_INCLUDED
#include <iostream>
#define IOSTREAM_INCLUDED
#endif

#ifndef UTILITY_INCLUDED
#include <utility>
#define UTILITY_INCLUDED
#endif

#ifndef ALGORITHM_INCLUDED
#include <algorithm>
#define ALGORITHM_INCLUDED
#endif

using namespace std;

void circumradius_inradius(float a, float b, float c, float &circumrad, float &inrad) {

    float s = (a+b+c) * 0.5;

    float den = s*(a+b-s)*(a+c-s)*(b+c-s);
    if (den<=0) {
	circumrad = 1;
	inrad = 0;
	return;
    }


    den = 4*sqrt(den);

    circumrad = (a*b*c) / den;
    inrad = (a*b*c) / (4*circumrad*s);
}

float face_quality(TriMesh *mesh, int face)
{
#define VERT(ix) mesh->vertices[mesh->faces[face][ix]]
	float l1 = len(VERT(1) - VERT(0)),
		l2 = len(VERT(2) - VERT(1)),
		l3 = len(VERT(0) - VERT(2)), c, i;

	circumradius_inradius(l1, l2, l3, c, i);
#undef VERT
	
	return i/c;
}

float vertex_quality(TriMesh *mesh, int vertex)
{
	float v = face_quality(mesh, mesh->adjacentfaces[vertex][0]);
	for (int i = 1; i < mesh->adjacentfaces[vertex].size(); ++i) {
		v = min(v, face_quality(mesh, mesh->adjacentfaces[vertex][i]));
	}
	return v;
}

int main(int argc, char **argv)
{
	TriMesh *mesh1 = TriMesh::read(argv[1]),
		*mesh2 = TriMesh::read(argv[2]),
		*mesh3 = TriMesh::read(argv[2]);
	mesh1->need_faces();
	mesh2->need_faces();
	mesh3->need_faces();

	mesh1->need_adjacentfaces();
	mesh2->need_adjacentfaces();
	mesh3->need_adjacentfaces();

	int count = mesh2->vertices.size();

	// true means using second mesh's vertices
	vector<bool> state(mesh3->vertices.size(), true);
	mesh3->colors = std::vector<Color>(mesh3->vertices.size(),
					   Color(0.2, 0.2, 1.0));

	bool changed;
	do {
	  changed = false;
		float current_quality = 10.0f;
		for (int j = 0; j < mesh3->vertices.size(); ++j) {
			point old_point = mesh3->vertices[j];
			float old_quality = vertex_quality(mesh3, j);
			switch (state[j]) {
			case false:
				mesh3->vertices[j] = mesh2->vertices[j];
				break;
			case true:
				mesh3->vertices[j] = mesh1->vertices[j];
				break;
			}
			float new_quality = vertex_quality(mesh3, j);
			if (new_quality <= old_quality) {
				mesh3->vertices[j] = old_point;
			} else {
			  if (state[j]) {
			    count--;
			    mesh3->colors[j] = Color(1.0, 0.2, 0.2);
			    changed = true;
			  } else {
			    count++;
			    mesh3->colors[j] = Color(0.2, 0.2, 1.0);
			    changed = true;
			  }
			  state[j] = !state[j];
			}
			current_quality = min(current_quality, 
					      max(old_quality, new_quality));
		}
		cout << "quality: " << current_quality << endl;
	} while (changed);

	mesh3->write(argv[3]);
	cout << "Total vertices: " << mesh1->vertices.size() << endl;
	cout << "Vertices chosen from mesh 1: " << mesh1->vertices.size() - count << endl;
	cout << "Vertices chosen from mesh 2: " << count << endl; 

	exit(0);	
}
