/* Copyright 2007 Universidade Federal do Rio Grande do Sul
   Carlos Dietrich - cadietrich@inf.ufrgs.br

   This copy is GPLv2 or later, your choice. */

#ifndef ERRORMANAGER_INCLUDED
#define ERRORMANAGER_INCLUDED

#include <vector>
#include <string>

// Mark an error at this line.
#define MarkError() CErrorManager::Instance().PushError(__FILE__, __LINE__)

//
#define ReportError(message) CErrorManager::Instance().PushError(__FILE__, __LINE__, message)

// This function is called by the windowing system when it wants to know if 
// there are any OpenGL errors.
#if defined(__gl_h_) || defined(__GL_H__)
#define ReportOpenGLErrors()\
{\
	GLenum error;\
	while ((error = glGetError()) != GL_NO_ERROR)\
	{\
		switch(error) {\
			case GL_INVALID_ENUM:\
				ReportError("Enum argument out of range.");\
				break;\
			case GL_INVALID_VALUE:\
				ReportError("Numeric argument out of range.");\
				break;\
			case GL_INVALID_OPERATION:\
				ReportError("Operation illegal in current state.");\
				break;\
			case GL_STACK_OVERFLOW:\
				ReportError("Command would cause a stack overflow.");\
				break;\
			case GL_STACK_UNDERFLOW:\
				ReportError("Command would cause a stack underflow.");\
				break;\
			case GL_OUT_OF_MEMORY:\
				ReportError("Not enough memory left to execute command.");\
				break;\
		}\
	}\
}
#endif // defined(__gl_h_) || defined(__GL_H__)

class CErrorManager
{
public:
	static CErrorManager& Instance(void)
	{
		if (!m_instance)
		{
			// Check for dead reference.
			if (m_isDestroyed)
				OnDeadReference();
			else
				// First callinitialize.
				Create();
		}

		return *m_instance;
	}
	void PushError(char *file, int line);
	void PushError(char *file, int line, const char *message);
	void PopError(void) const;
	void FlushErrors(void) const;
	bool IsOk(void) const;
private:
	CErrorManager(void);
	CErrorManager(const CErrorManager&);
	CErrorManager& operator=(const CErrorManager&);
	virtual ~CErrorManager(void);
	// Create a new CErrorManager and store a pointer to it in m_instance.
	static void Create(void);
	// Gets called if dead reference detected
	static void OnDeadReference(void);
	std::string FormatError(char *file, int line, const char *message) const;
protected:
	static std::vector<std::string> m_errorStack;
	static CErrorManager* m_instance;
	static bool m_isDestroyed;
};

#endif // ERRORMANAGER_INCLUDED

