/* Copyright 2007 Universidade Federal do Rio Grande do Sul
   Carlos Dietrich - cadietrich@inf.ufrgs.br

   This copy is GPLv2 or later, your choice. */

#ifndef IMARCHINGMETHOD_INCLUDED
#define IMARCHINGMETHOD_INCLUDED

#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>

#include <GageAdaptor.h>
#include <VertexPool.h>

class IMarchingMethod
	: boost::noncopyable
{
public:
	enum EDGE_TRANSFORM {
		NONE,
		GRADIENT_TRANSFORM,
		TANGENTIAL_TRANSFORM
	};
	enum MARCHING_METHOD {
		MARCHING_CUBES,
		MARCHING_TETRAHEDRA,
		DUAL_CONTOURING
	};
	IMarchingMethod(void);
	virtual ~IMarchingMethod(void);
	bool Load(boost::shared_ptr<CGageAdaptor>& image);
	virtual bool SetTransform(EDGE_TRANSFORM edgeTransform);
	virtual bool SetMethod(MARCHING_METHOD marchingMethod);
	virtual bool SetCellSize(float width, float height, float depth);
	void GetCellSize(float *width, float *height, float *depth) const;
	virtual bool SetGridSize(int width, int height, int depth);
	void GetGridSize(int *width, int *height, int *depth) const;
	virtual bool GenerateMesh(float isoLevel) = 0;
	float GetIsoLevel(void) const;
	boost::shared_ptr<CVertexPool> GetMesh(void) const;
	boost::shared_ptr<CVertexPool> GetActiveEdges(void) const;
	float GetAlgebraicError(void) const;
	bool IsValid(void) const;
protected:
	virtual void SetupVertexPool(void) = 0;
	void DeleteMesh(void);
	void TransformEdge(float *orgCoord, float *destCoord) const;
	void ApplyGradientTransform(float *orgCoord, float *destCoord) const;
	void ApplyTangentialTransform(float *orgCoord, float *destCoord) const;
	virtual int FindIntersection(float *orgCoord, float *destCoord);
private:
	void Create(void);
protected:
	boost::shared_ptr<CGageAdaptor> m_image;
	float m_cellWidth;
	float m_cellHeight;
	float m_cellDepth;
	int m_gridWidth;
	int m_gridHeight;
	int m_gridDepth;
	boost::shared_ptr<CVertexPool> m_triangleMesh;
	boost::shared_ptr<CVertexPool> m_activeEdges;
	EDGE_TRANSFORM m_edgeTransform;
	MARCHING_METHOD m_marchingMethod;
	float m_isoLevel;
#ifndef _NDEBUG // Debug only.
	float m_algebraicError;
#endif // _NDEBUG
};

#endif // IMARCHINGMETHOD_INCLUDED

