/* Copyright 2007 Universidade Federal do Rio Grande do Sul
   Carlos Dietrich - cadietrich@inf.ufrgs.br

   This copy is GPLv2 or later, your choice. */

#ifdef _WIN32
#include <windows.h>
#endif // _WIN32

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif


#include <iostream>
#include <fstream>

#include <ErrorManager.h>
#include <VertexPool.h>

using namespace std;

/**
*/
CVertexPool::CVertexPool(void)
{
	Create();
}

/**
*/
CVertexPool::~CVertexPool(void)
{
}

/**
*/
void CVertexPool::Clear(void)
{
	m_vertexArray.clear();
	m_colorArray.clear();
	m_normalArray.clear();
	m_texCoordArray.clear();

	m_elementArray.clear();
}

/**
*/
bool CVertexPool::Draw(void) const
{
	if (!m_colorArray.empty())
	{
		glColorPointer(m_colorSize, GL_FLOAT, 0, &m_colorArray[0]);
		
		glEnableClientState(GL_COLOR_ARRAY);
	}
	
	if (!m_normalArray.empty())
	{
		glNormalPointer(GL_FLOAT, 0, &m_normalArray[0]);

        glEnableClientState(GL_NORMAL_ARRAY);
	}

	if (!m_texCoordArray.empty())
	{
		glTexCoordPointer(m_texCoordSize, GL_FLOAT, 0, &m_texCoordArray[0]);

		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	}

	if (!m_vertexArray.empty())
	{
		glVertexPointer(m_vertexSize, GL_FLOAT, 0, &m_vertexArray[0]);

        glEnableClientState(GL_VERTEX_ARRAY);
	}
	
	if (!m_elementArray.empty())
		glDrawElements(m_mode, (GLsizei)m_elementArray.size(), GL_UNSIGNED_INT, &m_elementArray[0]);

	if (!m_colorArray.empty())
		glDisableClientState(GL_COLOR_ARRAY);
	
	if (!m_normalArray.empty())
        glDisableClientState(GL_NORMAL_ARRAY);

	if (!m_texCoordArray.empty())
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	if (!m_vertexArray.empty())
        glDisableClientState(GL_VERTEX_ARRAY);

	return true;
}

/**
*/
bool CVertexPool::Save(const string fileName, STREAM_FORMAT type, bool saveColor) const
{
	ofstream modelFile;
	string extendedFileName;
	const float *vertex;
	const float *color;
	const unsigned int *element;
	int vertexCount,
		elementCount,
		primitiveSize,
		primitiveCount,
		i;

	extendedFileName += fileName;

	vertex = &m_vertexArray[0];
	vertexCount = (int)m_vertexArray.size()/m_vertexSize;

	element = &m_elementArray[0];
	elementCount = (int)m_elementArray.size();
	
	if (saveColor) {
		color = &m_colorArray[0];
	}

	switch (m_mode) {
		case GL_POINTS:
			primitiveSize = 1;
			break;
		case GL_LINES:
			primitiveSize = 2;
			break;
		case GL_TRIANGLES:
			primitiveSize = 3;
			break;
		case GL_QUADS:
			primitiveSize = 4;
			break;
		case GL_LINE_LOOP:
		case GL_LINE_STRIP:
		case GL_TRIANGLE_STRIP:
		case GL_TRIANGLE_FAN:
		case GL_QUAD_STRIP:
		case GL_POLYGON:
			MarkError();

			return false;
			break;
	};

	primitiveCount = elementCount/primitiveSize;

	switch (type) {
		case OFF:
			modelFile.open(extendedFileName.c_str());

			modelFile << "OFF" << endl;
			modelFile << vertexCount << " " << primitiveCount << " " << primitiveCount*primitiveSize << endl;

			for ( ; vertexCount--; )
			{
				modelFile << *vertex << " " << *(vertex + 1) << " " << *(vertex + 2) << endl;

				vertex += 3;
			}

			for ( ; primitiveCount--; )
			{
				modelFile << primitiveSize << " ";

				for (i=primitiveSize; i--; )
					modelFile << *element++ << " ";

				modelFile << endl;
			}
			break;
		case PLY:
			modelFile.open(extendedFileName.c_str());

			modelFile << "ply" << endl;
			modelFile << "format ascii 1.0" << endl;
			modelFile << "element vertex " << vertexCount << endl;
			modelFile << "property float x" << endl;
			modelFile << "property float y" << endl;
			modelFile << "property float z" << endl;
			if (saveColor) {
				modelFile << "property uchar red" << endl;
				modelFile << "property uchar green" << endl;
				modelFile << "property uchar blue" << endl;
			}
			modelFile << "element face " << primitiveCount << endl;
			modelFile << "property list uchar int vertex_indices" << endl;
			modelFile << "end_header" << endl;

			for ( ; vertexCount--; )
			{
				if (saveColor) {
					modelFile << *vertex << " " << *(vertex + 1) << " " << *(vertex + 2) << " ";
					unsigned int red, green, blue;
					red = (unsigned int) (*color * 255.0f);
					green = (unsigned int) (*(color + 1) * 255.0f);
					blue = (unsigned int) (*(color + 2) * 255.0f);
					
					modelFile << red << " " << green << " " << blue << endl;
				}
				else
					modelFile << *vertex << " " << *(vertex + 1) << " " << *(vertex + 2) << endl;

				vertex += 3;
				if (saveColor) color+= 3;
			}

			for ( ; primitiveCount--; )
			{
				modelFile << primitiveSize << " ";

				for (i=primitiveSize; i--; )
					modelFile << *element++ << " ";

				modelFile << endl;
			}
			break;
	};

	return true;
}

/**
*/
void CVertexPool::Create(void)
{
	m_vertexSize = 0;
	m_vertexArray.clear();

	m_colorSize = 0;
	m_colorArray.clear();

	m_normalArray.clear();

	m_texCoordSize = 0;
	m_texCoordArray.clear();

	m_mode = GL_POINTS;

	m_elementArray.clear();
}

